import pytest
import os
import boto3
import jwt as pyjwt
import urllib.parse
import copy

from contextlib import contextmanager
from dli.client.builders import DatasetBuilder
from dli.client.dli_client import DliClient, Session

from tests.integration import random_name


def pytest_generate_tests(metafunc):
    metafunc.parametrize('client', ('QA',), indirect=['client'],
                         scope='function')
    metafunc.parametrize('jwt', ('QA',), indirect=['jwt'], scope='function')


class aws_envs:
    DEV = '116944431457'
    PROD = '867407613858'


@pytest.fixture
def jwt(request):
    # The env variable QA_OIDC_AUD can be found in gitlab under
    # the settings:CI/CD:Variables section and set locally.
    aud = os.environ[f'{request.param}_OIDC_AUD']
    return {
        "datalake": {
            "user_id": "3e1c688b-86f3-4b60-b988-15cb2d9b0133",
            "accounts": {
                "datalake-test1": "rw",
                "datalake-test2": "rw",
                "datalake-ops": "rw",
            },
            "groups": [],
            "organisation_id": os.environ['QA_ORG_ID']
        },
        "sub": "api:rollo-qa@ihsmarkit.com",
        "exp": 9564743138,
        "aud": aud
    }


class TestDliClient(DliClient):
    def __init__(self, api_root=None,
                 auth_key=None, host=None,
                 debug=False, jwt=None, secret=None,
                 access_id='noop', secret_key='noop'):
        self.jwt = jwt
        self.auth_key = auth_key
        self.secret = secret
        self.api_root = api_root
        super().__init__(api_root, host=host, debug=debug,
                         access_id=access_id,
                         secret_key=secret_key)

    def _new_session(self):
        return Session(
            None,
            self.secret_key,
            self._environment,
            self.host,
            self.auth_key,
        )

    @contextmanager
    def with_accounts(self, accounts):
        new_jwt = copy.deepcopy(self.jwt)
        new_jwt['datalake']['accounts'] = {
            account: 'rw' for account in accounts
        }

        auth_key = pyjwt.encode(new_jwt, self.secret).decode('utf8')
        yield TestDliClient(
            access_id=self.access_id,
            secret_key=self.secret_key,
            api_root=self.api_root,
            auth_key=auth_key
        )


@pytest.fixture
def client(monkeypatch, jwt, request):
    # This is the S3 address used for the QA environment.
    api_root = os.environ[f'{request.param}_API_URL']
    secret = os.environ[f'{request.param}_OIDC_SECRET']
    auth_key = pyjwt.encode(jwt, secret).decode('utf8')
    client = TestDliClient(
        api_root=api_root,
        auth_key=auth_key,
        secret=secret,
        jwt=jwt
    )
    yield client


@pytest.fixture
def aws_roles(client):
    """
    Ensures that the AWS roles are set up correctly. Sometimes
    QA messes with this setting.
    """
    aws_account = {
        'awsAccountId': aws_envs.PROD,
        'awsAccountName': 'dl_prod_aws_account',
        'awsRoleArn': 'arn:aws:iam::867407613858:role/trust-entity-dev-test',
        'awsRegion': 'eu-west-1',
        'accountIds': ['datalake-test1', 'datalake-test2', 'datalake-mgmt']
    }

    client.session.post(
        '__api/me/aws-accounts', json=aws_account
    )


@pytest.fixture
def package(client):
    package = client.register_package(
        name=random_name(),
        description="my package description",
        topic="Automotive",
        access="Restricted",
        internal_data="Yes",
        data_sensitivity="Public",
        terms_and_conditions="Terms",
        publisher="my publisher",
        access_manager_id='datalake-ops',
        tech_data_ops_id='datalake-test2',
        manager_id='datalake-test2'
    )
    yield package
    client.delete_package(package.package_id)


@pytest.fixture
def csv_dataset_builder(package):
    dataset_builder = DatasetBuilder(
        package_id=package.package_id,
        name='dataset-files-test-' + random_name(),
        description="My dataset description",
        content_type="Pricing",
        data_format='CSV',
        publishing_frequency="Weekly",
        taxonomy=[]
    )

    dataset_builder = dataset_builder.with_external_s3_storage(
        bucket_name='test-datalake-read-write-bucket-1',
        aws_account_id=aws_envs.PROD,
        prefix='abc',
    )
    yield dataset_builder


@pytest.fixture
def empty_dataset(client, aws_roles, csv_dataset_builder):
    dataset = client.register_dataset(csv_dataset_builder)
    yield dataset


@pytest.fixture
def csv_dataset(client, aws_roles, csv_dataset_builder):
    dataset = client.register_dataset(csv_dataset_builder)
    # NOTE register_s3_datafile both registers the datafile in the
    # datalake and it uploads the file to the bucket.
    client.register_s3_datafile(
        dataset_id=dataset.dataset_id,
        name='csv_dataset',
        files=['tests/integration/data/AAPL.csv'],
        s3_prefix='abc',
        data_as_of='2000-01-01'
    )
    yield dataset



import logging
import os

import warnings
import datetime
import glob
import s3fs
import multiprocessing

from concurrent.futures import ThreadPoolExecutor
from wrapt import ObjectProxy
from datetime import datetime, timedelta
from dateutil import tz
from boto3 import Session
import botocore
from urllib.parse import urlparse, ParseResult
from dli.client.exceptions import InsufficientPrivilegesException, S3FileDoesNotExist

logger = logging.getLogger(__name__)
THREADS = multiprocessing.cpu_count()


class Client:
    """
    A wrapper client providing util methods for s3fs
    """

    @classmethod
    def from_dataset(cls, client, dataset_id):
        def _credentials():
            keys = client.get_s3_access_keys_for_dataset(dataset_id)
            expiry_time = (
                # expiring the token after an hour. It's conservative
                # but it should just be refreshed. TODO add expiry time
                # to the get_s3_access_keys_for_dataset endpoint
                datetime.now().replace(tzinfo=tz.tzlocal()) + timedelta(hours=1)
            ).isoformat()

            return dict(
                access_key=keys.access_key_id,
                secret_key=keys.secret_access_key,
                token=keys.session_token,
                expiry_time=expiry_time
            )

        session_credentials = botocore.credentials.RefreshableCredentials.create_from_metadata(
            metadata=_credentials(),
            refresh_using=_credentials,
            method='NOOP'
        )

        session = botocore.session.get_session()
        # Extremely dirty but it seems to be the 'blessed' way of doing this.
        session._credentials = session_credentials
        return cls(Session(botocore_session=session))

    def __init__(self, boto_session):
        self.boto_session = boto_session

        self.s3_resource = self.boto_session.resource(
            's3', endpoint_url=os.environ.get('S3_URL')
        )

        self.s3_client = self.s3_resource.meta.client
        self.s3fs = s3fs.S3FileSystem()
        self.s3fs.s3 = self.s3_client

    def download_files_from_s3_path(self, s3_path, destination, flatten=False):
        """
        Helper function to download a file as a stream from S3
        :param str s3_path: required, example: s3://bucket_name/prefix_part1/prefix_part2/file_name.txt

        :param str destination: required. The path on the system, where the
            files should be saved. must be a directory, if doesn't exist, will
            be created.

        :param bool flatten: The default behaviour (=False) is to use the s3
            file structure when writing the downloaded files to disk. Example:
            [
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              'storm/climate/storm_data/storm_fatalities/as_of_date=2019-09-10/type=full/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

            When flatten = True, we remove the s3 structure. Example:

            Example output for new behaviour:
            [
              './storm-flattened/StormEvents_details-ftp_v1.0_d1950_c20170120.csv.gz',
              './storm-flattened/StormEvents_details-ftp_v1.0_d1951_c20160223.csv.gz'
            ]

        :returns: a list of the paths where the files were downloaded
            to e.g. ['path/git /package/dataset/as_of_date=2019-10-16
            /file.csv.gz'].
            with flatten=True e.g. ['path/file.csv.gz']

        :rtype: str
        """
        if os.path.isfile(destination):
            raise NotADirectoryError(destination)

        parse_result = urlparse(s3_path)
        bucket = parse_result.netloc
        # Removes forward slash
        download_path = parse_result.path.lstrip('/').rstrip('/')

        def _download(obj_summary):
            # Ignore 'folders'
            key = obj_summary['Key']
            destination_key = key
            if not key.endswith('/'):

                def flatten_s3_file_path(s3_path):
                    from urllib.parse import urlparse
                    import ntpath
                    path = urlparse(s3_path)[2]  # [scheme, netloc, path, ...]
                    head, tail = ntpath.split(path)
                    file = tail or ntpath.basename(head)
                    return file

                if flatten:
                    destination_key = flatten_s3_file_path(s3_path=key)

                destinatation_path = os.path.join(destination, destination_key)
                subfolder = os.path.dirname(destinatation_path)
                os.makedirs(subfolder, exist_ok=True)
                logger.info('Downloading File %s to %s', key, destinatation_path)
                self.s3_client.download_file(bucket, key, destinatation_path)
                return destinatation_path

        keys = list(self._list_objects(bucket, download_path))
        threads = min(THREADS, len(keys))
        with ThreadPoolExecutor(max_workers=threads) as executor:
            destination_path = list(executor.map(_download, keys))
        return destination_path

    def upload_files_to_s3(self, files, s3_location, token_refresher=None, num_retries=None):
        """
        Upload multiple files to a specified s3 location. The basename of the
        `files` will be preserved.

        :param List files: An list of filepaths
        :param str s3_location: Path to destination directory in S3, file will be
                 stored under <s3_location><filename>
        :param token_refresher: Optional Function to refresh S3 token *deprecated*
        :param int num_retries: Optional Number of retries in case of upload failure *deprecated*
        :returns: List of path to the files in S3
        :rtype: List[str]
        """
        if token_refresher:
            warnings.warn(
                'Credentials are now automatically refreshed.'
                'This will be removed in the next major release.',
                DeprecationWarning,
            )

        if num_retries:
            warnings.warn(
                'Retries are handled by boto.'
                'This will be removed in the next major release.',
                DeprecationWarning,
            )


        bucket_name, *prefix = s3_location.partition('/')
        bucket = self.s3_resource.Bucket(bucket_name)
        prefix = ''.join(prefix)


        def upload(file, s3_location):
            file_path = file['file']
            s3_suffix = file['s3_suffix']
            target = os.path.join(prefix, s3_suffix)

            logger.info("Uploading %s to %s", file_path, target)

            try:
                self.s3_client.upload_file(
                    file_path, bucket.name, target.lstrip('/')
                )
            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == 'AccessDenied':
                    raise InsufficientPrivilegesException() from e

                raise e

            logger.info(".. %s uploaded successfully.", file_path)

            return {
                "path": ParseResult(
                    scheme='s3',
                    netloc=bucket_name,
                    path=target,
                    params='', query='', fragment=''
                ).geturl(),
                "size": os.path.getsize(file_path)
            }

        files_to_upload = self._files_to_upload_flattened(files)
        result = []
        for file in files_to_upload:
            uploaded = upload(file, s3_location)

            result.append(uploaded)

        return result

    @staticmethod
    def _files_to_upload_flattened(files):
        files_to_upload = []

        for file in files:
            if not os.path.exists(file):
                raise Exception('File / directory specified (%s) for upload does not exist.' % file)

            if os.path.isfile(file):
                logger.info("detected file: %s", file)
                files_to_upload.append({'file': file,
                                        's3_suffix': os.path.basename(file)})
            elif os.path.isdir(file):
                logger.info("detected directory: %s", file)
                all_contents = glob.glob(os.path.join(file, "**/*"), recursive=True)
                fs = [f for f in all_contents if os.path.isfile(f)]

                for f in fs:
                    files_to_upload.append({
                        'file': f,
                        's3_suffix': os.path.relpath(f, file).replace("\\", "/")
                    })

        return files_to_upload

    def _list_objects(self, bucket, prefix):
        # as it turns out the high level boto client is rather expensive
        # when creating objects en mess. This is at least 50% faster.
        # This is also the exact way that the aws cli lists files.
        paginator = self.s3_client.get_paginator(
            'list_objects_v2'
        )

        paging_args = {'Bucket': bucket, 'Prefix': prefix}
        iterator = paginator.paginate(**paging_args)

        for response_data in iterator:
            for obj in response_data.get('Contents'):
                yield obj


class File(ObjectProxy):
    """
    This is only to mantain backwards compatibility with S3FS.
    Ideally we want to deprecate getting files as anything but
    context managers. This provides both the context manageer
    and function that returns a files like object
    """
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.__wrapped__.close()


class S3DatafileWrapper:
    def __init__(self, datafile, client):
        # we want to override the files property
        # it isn't really needed for py3 but it is in py2
        self.__dict__ = {
            k: v for k,v in datafile.items() if k != "files"
        }
        self._datafile = datafile
        self._client = client

    def __getitem__(self, path):
        """
        :param str path: The path of the file to retrieve from s3.
        :returns: A S3File instance
        """
        return self.open_file(path)

    def open_file(self, path):
        """
        Helper method to load an specific file inside a datafile
        without having to download all of it.

        :param str path: The path of the file to retrieve from s3.
        :returns: A S3File instance that can be used as a normal python file
        """
        bucket, _, key = path.partition('/')

        try:
            bucket = self._client.s3_resource.Bucket(bucket)
            response = bucket.Object(key).get()['Body']
            return File(response)
        except (
            self._client.s3_resource.meta.client.exceptions.NoSuchKey,
            self._client.s3_resource.meta.client.exceptions.NoSuchBucket
        ) as e:
            raise S3FileDoesNotExist(path)

    def __repr__(self):
        return str(dict(self._datafile))

    @property
    def files(self):
        """
        Lists all S3 files in this datafile, recursing on folders (if any)
        """
        files = self._datafile["files"]
        if not files:
            return []

        bucket = urlparse(files[0]['path']).netloc

        prefix = os.path.commonprefix([
            urlparse(file_['path']).path for file_ in files
        ]).lstrip('/')

        return [
            '{}/{}'.format(bucket, obj['Key']) for obj in
            self._client._list_objects(bucket, prefix)
        ]



from unittest.mock import patch, MagicMock

import pytest
from dli.client.components.urls import package_urls


class TestPackageModel:
    def test_display_package(
        self, capsys, dataset_request_index, test_client,
        package_request_v2
    ):
        test_client._session.get.return_value.json.side_effect = [
            dataset_request_index
        ]

        package = test_client._Package(package_request_v2)

        package.contents()
        captured = capsys.readouterr()
        assert captured.out == (
            '\nDATASET "TestDataset" [PARQUET]\n'
            '>> Shortcode: TestDataset\n'
            '>> Available Date Range: 2019-05-06 to 2020-01-01\n'
            '>> ID: 5b01376e-975d-11e9-8832-7e5ef76d533f\n'
            '>> Published: Monthly by IHS Markit\n'
            '>> Accessible: False\n'
            '\ndescription'
            '\n---------------------------------------'
            '-----------------------------------------\n'
        )

    @pytest.mark.xfail(
        reason='The packages() and datasets() endpoints are disabled')
    def test_retrieve_datasets_given_package(self, test_client,
                                             package_request, package_request_v2,
                                             dataset_request_index):
        test_client._session.get.return_value.json.side_effect = [
            package_request, package_request_v2, dataset_request_index
        ]

        pp = test_client.packages(search_term="", only_mine=False)
        ds = pp["Test Package"].datasets()
        assert len(ds) == 1


class TestPackageModule:

    @pytest.mark.xfail(
        reason='The packages() and datasets() endpoints are disabled')
    def test_search_packages(
        self, test_client, package_request, dataset_request_index
    ):
        with patch('dli.modules.package.Paginator') as mock_paginator:
            test_client.packages(search_term='abc')
            assert (
                mock_paginator.call_args_list[0][0][0] ==
                '/__api/search/packages/?query=abc'
            )

            assert (
                mock_paginator.call_args_list[1][0][0] ==
                '/__api/me/consumed_packages/'
            )

            mock_paginator.reset_mock()

            test_client.packages(search_term='abc', only_mine=False)
            assert mock_paginator.call_count == 1
            assert (
                mock_paginator.call_args_list[0][0][0] ==
                '/__api/search/packages/?query=abc'
            )

    def test_retrieve_package_by_get(
        self, test_client, package_request, dataset_request_index
    ):
        name = package_request['entities'][0]['properties']['name']
        package_request_v2 = {
            'data': {
                'id': package_request['entities'][0]['properties']['packageId'],
                'attributes': {'name': name}
            }
        }

        test_client._session.get.return_value.json.side_effect = [
                MagicMock(), package_request_v2
        ]

        test_package = test_client.packages._get('Test Package')
        assert test_package.name == name



import urllib


class dataset_urls:
    instance = '/__api/datasets/{id}/'
    index = '/__api/datasets/'
    datafiles = '/__api/datasets/{id}/datafiles/'
    latest_datafile = '/__api/datasets/{id}/latest-datafile/'
    keys = '/__api/datasets/{id}/keys/'
    access_keys = '/__api/request-access-keys/'

    # dli_v2
    v2_index = '/__api_v2/datasets/'
    v2_instance = '/__api_v2/datasets/{id}/'
    v2_by_short_code = '/__api_v2/by_short_code/dataset/{dataset_short_code}'
    v2_sample_data_schema = '/__api_v2/datasets/{id}/sample_data'
    v2_sample_data_file = '/__api_v2/datasets/{id}/sample_data/file'
    v2_schema_instance_version = '/__api_v2/datasets/{id}/dictionaries/{version}'

    schemas_many = '/__api/datasets/{id}/schemas/'
    schema_instance = '/__api/datasets/{id}/schema/'

    # V2 API
    dictionary_index =  '/__api_v2/dictionaries/'
    dictionary_instance = '/__api_v2/dictionaries/{id}/'
    dictionary_by_dataset = '/__api_v2/datasets/{id}/dictionaries/'
    dictionary_by_dataset_lastest = '/__api_v2/datasets/{id}/current_dictionary/'
    dictionary_fields = '/__api_v2/dictionaries/{id}/fields/'


class autoreg_urls:
    autoreg_index = '/__api/auto-reg-metadata/'
    autoreg_instance = '/__api/auto-reg-metadata/{id}/'


class datafile_urls:
    datafiles_index = '/__api/datafiles/'
    datafiles_instance = '/__api/datafiles/{id}/'


class me_urls:
    my_packages = '/__api/me/my-packages/'
    consumed_packages = '/__api/me/consumed_packages/'
    my_accounts = '/__api/me/my_accounts/'


class package_urls:
    package_index = '/__api/packages/'
    package_edit = '/__api/packages/{id}/'
    # For some reason these are 'package' not 'package*s*'
    package_datasets = '/__api/package/{id}/datasets/'
    v2_package_datasets = '/__api_v2/packages/{id}/datasets'
    v2_package_index = '/__api_v2/packages/{id}'


class search_urls:
    search_root = '/__api/search/'
    search_packages = '/__api/search/packages/'
    search_datasets = '/__api/search/datasets/'


class consumption_urls:
    consumption_download = '/datafile/{id}/download/binary/{path}'
    consumption_manifest = '/datafile/{id}/manifest/'
    consumption_dataframe = '/dataset/{id}/dataframe/'
    consumption_analytics = '/analytics/'
    consumption_partitions = '/dataset/{id}/partitions/'


class accounts_urls:
    account_instance = '/__api/account/{id}/'

class sam_urls:
    sam_token = '/sso/oauth2/realms/root/realms/Customers/access_token'

class identity_urls:
    identity_token = '/api/identity/v2/auth/token'
